from .TokenSpan import TokenSpan


class Sentence(TokenSpan):
	def __init__(self, obj, document):
		super().__init__(obj=obj, document=document)

	@property
	def id(self):
		return self.document.id, 'sentence', self.start, self.end



import numpy as np

from .get_string_similarity import get_string_similarity

def get_sentence_distance(words1, words2, first_char_weight=0, case_sensitivity=1, method='jaro_winkler'):
	"""
	:type words1: list[str] or str
	:type words2: list[str] or str
	:type first_char_weight: float
	:type case_sensitivity: float
	:type method: str
	:rtype: float
	"""
	if isinstance(words1, str):
		words1 = words1.split()
	elif not isinstance(words1, list):
		raise TypeError('words1 should either be a string or a list!')

	if isinstance(words2, str):
		words2 = words2.split()
	elif not isinstance(words2, list):
		raise TypeError('words2 should either be a string or a list')

	size_x = len(words1) + 1
	size_y = len(words2) + 1
	m = np.zeros((size_x, size_y))
	# create_matrix(h=size_x, w=size_y, obj=0)

	for i in range(size_x): m[i, 0] = i
	for j in range(size_y): m[0, j] = j
	for i in range(1, size_x):
		for j in range(1, size_y):
			w1 = words1[i - 1]
			w2 = words2[j - 1]
			normalized_distance = 1 - get_string_similarity(
				s1=w1, s2=w2, method=method,
				first_char_weight=first_char_weight,
				case_sensitivity=case_sensitivity
			)
			m[i][j] = min(
				m[i - 1, j - 1] + normalized_distance,
				m[i - 1, j] + 1,
				m[i, j - 1] + 1
			)

	return m[size_x - 1, size_y - 1]


def get_sentence_similarity(
		words1, words2, first_char_weight=0, first_word_weight=0, case_sensitivity=1, method='jaro_winkler'
):
	"""
	:type words1: list[str] or str
	:type words2: list[str] or str
	:type first_char_weight: float
	:type first_word_weight: float
	:type case_sensitivity: float
	:type method: str
	:rtype: float
	"""
	if isinstance(words1, str):
		words1 = words1.split()
	elif not isinstance(words1, list):
		raise TypeError('words1 should either be a string or a list!')

	if isinstance(words2, str):
		words2 = words2.split()
	elif not isinstance(words2, list):
		raise TypeError('words2 should either be a string or a list')

	distance = get_sentence_distance(
		words1=words1, words2=words2, method=method,
		case_sensitivity=case_sensitivity, first_char_weight=first_char_weight
	)
	if len(words1)==0 and len(words2)==0:
		return 1.0
	elif len(words1)==0 or len(words2)==0:
		return 0.0

	similarity = 1 - distance/max(len(words1), len(words2))

	if first_word_weight==0:
		return similarity
	elif first_word_weight<0:
		raise ValueError('first_word_weight cannot be negative!')
	else:
		first_word_similarity = get_string_similarity(
			s1=words1[0], s2=words2[0], method=method,
			case_sensitivity=case_sensitivity, first_char_weight=first_char_weight
		)
		return (similarity+first_word_similarity*first_word_weight)/(1+first_word_weight)




import requests
from ..document.part_of_speach_definitions import PART_OF_SPEACH

_analyze_to_engine = {
	'entities': 'analyzeEntities',
	'entity_sentiment': 'analyzeEntitySentiment',
	'sentiment': 'analyzeSentiment',
	'syntax': 'analyzeSyntax',
	'all': 'annotateText',
	'document': 'classifyText'
}


def standardize_text(text):
	if 'beginOffset' in text:
		text['begin_offset'] = text.pop('beginOffset')


def standardize_token(token):
	if 'text' in token:
		text = token['text']
		standardize_text(text=text)

	if 'partOfSpeech' in token:
		part_of_speech = token.pop('partOfSpeech')
		token['part_of_speech'] = {
			key: standardirze_part_of_speech(value.lower()) if key == 'tag' else value.lower()
			for key, value in part_of_speech.items()
			if not value.endswith('_UNKNOWN')
		}

	if 'dependencyEdge' in token:
		dependency_edge = token.pop('dependencyEdge')
		if 'headTokenIndex' in dependency_edge:
			dependency_edge['head_token_index'] = dependency_edge.pop('headTokenIndex')
		if 'label' in dependency_edge:
			dependency_edge['label'] = dependency_edge['label'].lower()


def standardize_entity(entity):
	if 'type' in entity:
		entity['type'] = entity['type'].lower()
	if 'mentions' in entity:
		mentions = entity['mentions']
		for mention in mentions:
			if 'type' in mention:
				mention['type'] = mention['type'].lower()
			if 'text' in mention:
				text = mention['text']
				standardize_text(text=text)


def standardirze_part_of_speech(part_of_speech):
	if part_of_speech in PART_OF_SPEACH:
		return PART_OF_SPEACH[part_of_speech]
	else:
		return part_of_speech


def analyze(
		text, api_key, standardize=True, language='en',
		extract_entities=True, extract_sentiment=True, extract_syntax=True
):
	document = {
		"document": {
			"type": "PLAIN_TEXT",
			"language": language,
			"content": text
		},
		'features': {
			"extractEntities": extract_entities,
			"extractDocumentSentiment": extract_sentiment,
			"extractEntitySentiment": extract_entities and extract_sentiment,
			"extractSyntax": extract_syntax
		},
		"encodingType": "UTF8"
	}
	response = requests.post(
		f"https://language.googleapis.com/v1beta2/documents:annotateText?key={api_key}",
		json=document, verify=False
	)
	result = response.json()
	if standardize:
		if 'sentences' in result:
			sentences = result['sentences']
			for sentence in sentences:
				if 'text' in sentence:
					text = sentence['text']
					standardize_text(text=text)
		if 'tokens' in result:
			tokens = result['tokens']
			for token in tokens:
				standardize_token(token=token)
		if 'documentSentiment' in result:
			result['document_sentiment'] = result.pop('documentSentiment')
		if 'entities' in result:
			for entity in result['entities']:
				standardize_entity(entity=entity)
	result['text'] = text
	return result





from .convert_score_and_distance_functions import convert_distance_function_to_score_function
from .convert_score_and_distance_functions import create_weighted_similarity_function

from .Word import Word
from .Sentence import Sentence
from .get_char_by_char_equality import get_char_by_char_equality

from .get_similarity import get_similarities, get_similarity
from .get_sentence_similarity import get_sentence_distance
from .get_string_similarity import get_string_similarity

from .find_most_similar import find_most_similar




from .GoogleDocumentObject import GoogleDocumentObject
from .GoogleToken import GoogleToken


class GoogleTokenSpan(GoogleDocumentObject):
	def __init__(self, dictionary, document, begin, end):
		super().__init__(dictionary=dictionary, document=document, begin=begin, end=end)
		self._tokens = None

	def __str__(self):
		return ', '.join([str(token) for token in self.tokens])

	def __repr__(self):
		return str(self)

	@property
	def tokens(self):
		"""
		:rtype: list[GoogleToken]
		"""
		if self._tokens is None:
			self._tokens = [
				token for token in self.document.tokens if token.begin >= self.begin and token.end <= self.end
			]
		return self._tokens



from jellyfish import jaro_winkler as _get_jaro_winkler_similarity


def get_jaro_winkler_similarity(s1, s2, case_sensitive=True):
	if not case_sensitive:
		s1 = s1.lower()
		s2 = s2.lower()
	return _get_jaro_winkler_similarity(s1, s2)


def get_jaro_winkler_distance(s1, s2, case_sensitive=True):
	return max(len(s1), len(s2))*(1-get_jaro_winkler_similarity(s1, s2, case_sensitive=case_sensitive))




# source: http://snowball.tartarus.org/algorithms/english/stop.txt

STOP_WORDS = {
	'a', 'about', 'above', 'across', 'after', 'afterwards', 'again', 'against', 'all', 'almost', 'alone', 'along',
	'already', 'also', 'although', 'always', 'am', 'among', 'amongst', 'amount', 'an', 'and', 'another', 'any',
	'anyhow', 'anyone', 'anything', 'anyway', 'anywhere', 'are', 'around', 'as', 'at', 'back', 'be', 'became',
	'because', 'become', 'becomes', 'becoming', 'been', 'before', 'beforehand', 'behind', 'being', 'below', 'beside',
	'besides', 'between', 'beyond', 'both', 'bottom', 'but', 'by', 'ca', 'call', 'can', 'cannot', 'could', 'did',
	'do', 'does', 'doing', 'done', 'down', 'due', 'during', 'each', 'eight', 'either', 'eleven', 'else', 'elsewhere',
	'empty', 'enough', 'even', 'ever', 'every', 'everyone', 'everything', 'everywhere', 'except', 'few', 'fifteen',
	'fifty', 'first', 'five', 'for', 'former', 'formerly', 'forty', 'four', 'from', 'front', 'full', 'further', 'get',
	'give', 'go', 'had', 'has', 'have', 'he', 'hence', 'her', 'here', 'hereafter', 'hereby', 'herein', 'hereupon',
	'hers', 'herself', 'him', 'himself', 'his', 'how', 'however', 'hundred', 'i', 'if', 'in', 'indeed', 'into', 'is',
	'it', 'its', 'itself', 'just', 'keep', 'last', 'latter', 'latterly', 'least', 'less', 'made', 'make', 'many',
	'may', 'me', 'meanwhile', 'might', 'mine', 'more', 'moreover', 'most', 'mostly', 'move', 'much', 'must', 'my',
	'myself', 'name', 'namely', 'neither', 'never', 'nevertheless', 'next', 'nine', 'no', 'nobody', 'none', 'noone',
	'nor', 'not', 'nothing', 'now', 'nowhere', 'of', 'off', 'often', 'on', 'once', 'one', 'only', 'onto', 'or',
	'other', 'others', 'otherwise', 'our', 'ours', 'ourselves', 'out', 'over', 'own', 'part', 'per', 'perhaps',
	'please', 'put', 'quite', 'rather', 're', 'really', 'regarding', 'same', 'say', 'see', 'seem', 'seemed',
	'seeming', 'seems', 'serious', 'several', 'she', 'should', 'show', 'side', 'since', 'six', 'sixty', 'so', 'some',
	'somehow', 'someone', 'something', 'sometime', 'sometimes', 'somewhere', 'still', 'such', 'take', 'ten', 'than',
	'that', 'the', 'their', 'them', 'themselves', 'then', 'thence', 'there', 'thereafter', 'thereby', 'therefore',
	'therein', 'thereupon', 'these', 'they', 'third', 'this', 'those', 'though', 'three', 'through', 'throughout',
	'thru', 'thus', 'to', 'together', 'too', 'top', 'toward', 'towards', 'twelve', 'twenty', 'two', 'under', 'unless',
	'until', 'up', 'upon', 'us', 'used', 'using', 'various', 'very', 'via', 'was', 'we', 'well', 'were', 'what',
	'whatever', 'when', 'whence', 'whenever', 'where', 'whereafter', 'whereas', 'whereby', 'wherein', 'whereupon',
	'wherever', 'whether', 'which', 'while', 'whither', 'who', 'whoever', 'whole', 'whom', 'whose', 'why', 'will',
	'with', 'within', 'without', 'would', 'yet', 'you', 'your', 'yours', 'yourself', 'yourselves'
}


PRONOUND_STOP_WORDS = {
	'i', 'me', 'my', 'myself', 'we', 'us', 'our', 'ours', 'ourselves',
	'you', 'your', 'yours', 'yourself', 'yourselves',
	'he', 'him', 'his', 'himself', 'she', 'her', 'hers', 'herself',
	'it', 'its', 'itself', 'they', 'them', 'their', 'theirs', 'themselves',
	'what', 'which', 'who', 'whom', 'this', 'that', 'these', 'those'
}


def get_stop_words(include_pronoun_forms=True):
	if include_pronoun_forms:
		return STOP_WORDS
	else:
		return STOP_WORDS - PRONOUND_STOP_WORDS
